@@ -50,7 +50,7 @@ from hypothesis.internal.conjecture.providers import (
 )
 from hypothesis.internal.floats import SIGNALING_NAN, clamp
 from hypothesis.internal.intervalsets import IntervalSet
-from hypothesis.internal.observability import TESTCASE_CALLBACKS, Observation
+from hypothesis.internal.observability import Observation, _callbacks
 
 from tests.common.debug import minimal
 from tests.common.utils import (
@@ -786,7 +786,7 @@ def test_on_observation_no_override():
     @given(st.integers())
     @settings(backend="observation")
     def f(n):
-        assert TESTCASE_CALLBACKS == []
+        assert _callbacks == {}
 
     f()
 
